/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.cluster;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.AbstractExecutorService;
import java.util.concurrent.BlockingDeque;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.log4j.Logger;
import org.jgroups.Address;
import org.jgroups.ChannelException;
import org.jgroups.JChannel;
import org.jgroups.Message;
import org.jgroups.ReceiverAdapter;
import org.jgroups.View;
import org.jgroups.util.Streamable;
import org.jgroups.util.Util;

import com.koutra.dist.proc.cluster.Request.Type;

/**
 * An executor service implementation that knows how to execute tasks on the JGroup
 * cluster.
 * <p>
 * This class and <code>ClusterNode</code> instances implement a simpler modified version
 * of the protocol specified in the
 * <a href="http://www.jgroups.org/taskdistribution.html">JGroups site</a>.
 * 
 * @author Pafsanias Ftakas
 */
public class ClusterExecutorService extends AbstractExecutorService
		implements ExecutorService {
	private static final Logger logger =
		Logger.getLogger(ClusterExecutorService.class);
	
	protected JChannel channel;
	protected AtomicInteger rank;
	protected AtomicInteger clusterSize;
	protected ClusterNode node;
	protected Map<Integer, ClusterNodeInfo> rankToNodeInfoMap;
	protected BlockingDeque<Request> pendingRequests;
	protected Map<TaskID, ClusterFuture> taskToFutureMap;
	
	/**
	 * Initializing constructor. This will actually connect to the cluster with two nodes:
	 * one for the executor in order that it will be able to submit tasks to the cluster,
	 * and one normal cluster node so that if nothing else exists, at least it will be
	 * available as a slave worker.
	 * @throws ChannelException iff something goes wrong when connecting to the cluster.
	 */
	public ClusterExecutorService() throws ChannelException {
		rank = new AtomicInteger(0);
		clusterSize = new AtomicInteger(0);
		rankToNodeInfoMap = new HashMap<Integer, ClusterNodeInfo>();
		// TODO In real world environments the deque should have a capacity limitation.
		pendingRequests = new LinkedBlockingDeque<Request>();
		taskToFutureMap = new HashMap<TaskID, ClusterFuture>();
		channel = new JChannel("udp.xml");
		channel.setReceiver(new ReceiverAdapter() {
			@Override
			public void receive(Message msg) {
				Request req = null;
				try {
					req = (Request) Util.objectFromByteBuffer(msg.getBuffer());
				} catch (Exception e) {
					logger.error("Unable to convert received message to request", e);
					return;
				}
				if (logger.isTraceEnabled())
					logger.trace("Cluster executor service received message " + req.getType() +
							" (" + req.getTaskID() + ")");
				
				switch (req.getType()) {
				case EXECUTE:
					logger.trace("Ignoring execute request.");
					break;
				case RESULT:
					logger.trace("Received response.");
					TaskID taskID = req.getTaskID();
					ClusterFuture future = taskToFutureMap.get(taskID);
					Request originalReq = future.getRequest();
					pendingRequests.remove(originalReq);
					taskToFutureMap.remove(taskID);
					future.setResult(req.getResult());
					break;
				default:
					logger.trace("Unknown request type. Ignoring.");
				}
			}
			
			@Override
			public void viewAccepted(View new_view) {
				logger.info("Cluster view has changed: " + new_view);
				Address localAddress = channel.getLocalAddress();
				Vector<Address> addresses = new_view.getMembers();
				Map<Integer, ClusterNodeInfo> lRankToNodeInfoMap =
					new HashMap<Integer, ClusterNodeInfo>();
				int new_rank = -1;
				for (int i = 0; i < addresses.size(); i++) {
					ClusterNodeInfo info = findNodeInfoFor(addresses.get(i));
					if (info == null)
						info = new ClusterNodeInfo(addresses.get(i), i);
					lRankToNodeInfoMap.put(i, info);
					if (localAddress.equals(addresses.get(i))) {
						new_rank = i;
					}
				}
				if (new_rank == -1)
					throw new RuntimeException("We were unable to find our node in " +
							"the view. Something is seriously wrong.");
				for (ClusterNodeInfo info : rankToNodeInfoMap.values()) {
					boolean found = false;
					for (ClusterNodeInfo lInfo : lRankToNodeInfoMap.values()) {
						if (lInfo.getAddress().equals(info.getAddress())) {
							found = true;
							break;
						}
					}
					if (!found) {
						// The cluster node has gone down. Panic! There is no recovery
						// that can take place, as there are consumable streams involved.
						logger.warn("Node " + info.getAddress() + " has gone down. Any " +
								"currently running pipelines will most probably fail.");
					}
				}
				rank.set(new_rank);
				clusterSize.set(new_view.size());
				rankToNodeInfoMap = lRankToNodeInfoMap;
				
				if (logger.isTraceEnabled()) {
					logger.trace("Cluster size is now: " + new_view.size() +
							", my rank is: " + new_rank);
				}
			}
		});
		channel.connect("DistProcCluster");
		node = new ClusterNode();
	}
	
	public ClusterNodeInfo findNodeInfoFor(Address addr) {
		for (ClusterNodeInfo info : rankToNodeInfoMap.values()) {
			if (info.getAddress().equals(addr))
				return info;
		}
		
		return null;
	}
	
	@Override
	public boolean awaitTermination(long timeout, TimeUnit unit)
			throws InterruptedException {
		// Silly implementation, but this is not really used, so I am not
		// spending too much time and effort to get this right.
		if (channel == null)
			return true;
		unit.sleep(timeout);
		return channel == null;
	}

	@Override
	public boolean isShutdown() {
		return channel == null;
	}

	@Override
	public boolean isTerminated() {
		return channel == null;
	}

	@Override
	public void shutdown() {
		try {
			node.shutdown();
			channel.disconnect();
			channel.close();
		} finally {
			node = null;
			channel = null;
		}
	}

	@Override
	public List<Runnable> shutdownNow() {
		shutdown();
		return Collections.emptyList();
	}

	/**
	 * Execute the parameter runnable in the cluster.
	 * @param command the command to execute. Must be an instance of
	 * <code>IClusterRunnable</code>, otherwise an <code>IllegalArgumentException</code>
	 * is thrown.
	 */
	@Override
	public void execute(Runnable command) {
		if (!(command instanceof IClusterRunnable))
			throw new IllegalArgumentException(
					"Argument to the ClusterExecutor class must be extending the " +
					"ClusterRunnable class. Class is: " +
					command.getClass().getCanonicalName());
		
		IClusterRunnable actualCommand = (IClusterRunnable) command;
		Streamable payload = actualCommand.getWirePayload();
		TaskID taskID = TaskID.createTaskID(rank.get(), clusterSize.get(),
				channel.getLocalAddress());
		Request req = new Request(Type.EXECUTE, payload, taskID, null);
		pendingRequests.add(req);
		try {
			Message msg = new Message(null, null, Util.objectToByteBuffer(req));
			channel.send(msg);
		} catch (Exception e) {
			logger.error("Unable to send message to cluster", e);
		}
		
		if (logger.isTraceEnabled())
			logger.trace("Adding task in taskToFutureMap for: " + taskID);

		ClusterFuture future = new ClusterFuture(req);
		taskToFutureMap.put(taskID, future);
	}
	
	@Override
	public Future<?> submit(Runnable command) {
		// We must return a future that is aware of the cluster and gets updated, as
		// task states get updated by the cluster nodes.
		
		if (!(command instanceof IClusterRunnable))
			throw new IllegalArgumentException(
					"Argument to the ClusterExecutor class must be extending the " +
					"ClusterRunnable class. Class is: " +
					command.getClass().getCanonicalName());
		
		IClusterRunnable actualCommand = (IClusterRunnable) command;
		Streamable payload = actualCommand.getWirePayload();
		TaskID taskID = TaskID.createTaskID(rank.get(), clusterSize.get(),
				channel.getLocalAddress());
		Request req = new Request(Type.EXECUTE, payload, taskID, null);
		pendingRequests.add(req);
		try {
			Message msg = new Message(null, null, Util.objectToByteBuffer(req));
			channel.send(msg);
		} catch (Exception e) {
			logger.error("Unable to send message to cluster", e);
		}
		
		if (logger.isTraceEnabled())
			logger.trace("Adding task in taskToFutureMap for: " + taskID);

		ClusterFuture future = new ClusterFuture(req);
		taskToFutureMap.put(taskID, future);
		return future;
	}
	
}
