/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.faucet;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItem;
import com.koutra.dist.proc.model.XformationException;

/**
 * A faucet implementation that reads using the java.io.Reader interface. The client
 * can optionally provide an input character set (if null, the system character set
 * will be used as specified by the java.io.FileReader documentation). When using
 * the constructor that supplies a java.io.Reader, the caller has the responsibility
 * of closing the reader once the transformation has completed.
 * 
 * @author Pafsanias Ftakas
 */
public class ReaderFaucet implements IFaucet {
	private static final Logger logger = Logger.getLogger(ReaderFaucet.class);
	
	protected enum Type {
		File,
		Reader;
	};

	protected String id;
	protected ISink sink;
	protected boolean hookedUp;
	protected Type type;
	protected Reader reader;
	protected String path;
	protected String inputCharset;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public ReaderFaucet() { }
	
	public ReaderFaucet(String id, Reader reader) {
		this.id = id;
		this.sink = null;
		this.hookedUp = false;
		this.type = Type.Reader;
		this.reader = reader;
		this.path = null;
		this.inputCharset = null;
	}

	public ReaderFaucet(String id, String path, String inputCharset) {
		this.id = id;
		this.sink = null;
		this.hookedUp = false;
		this.type = Type.File;
		this.reader = null;
		this.path = path;
		this.inputCharset = inputCharset;
	}

	@Override
	public String getId() {
		return id;
	}
	
	@Override
	public ISink getSink() {
		return sink;
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case CharStream:
			return true;
		case ByteStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public void hookupSink(ISink sink) {
		if (!sink.supportsInput(ContentType.CharStream))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the CharStream content type.");
		if (sink instanceof IMuxSink)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an IMuxSink instance. This is not allowed.");
		if (sink instanceof ISinkTemplate)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an ISinkTemplate instance. This is not allowed.");
		if (hookedUp && this.sink == sink) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUp)
			throw new XformationException("Trying to hook up an already hooked up faucet.");
		
		this.sink = sink;
		this.hookedUp = true;
		
		this.sink.hookupFaucet(this);
	}
	
	@Override
	public Object getSource(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
		case XML:
		case ResultSet:
			return null;
		}

		if (reader != null)
			return reader;
		
		switch (type) {
		case File:
			try {
				if (inputCharset == null) {
					FileReader r = new FileReader(path);
					reader = new BufferedReader(r);
					
					if (logger.isTraceEnabled())
						logger.trace("Faucet has created reader " + reader +
								" wrapping reader " + r + " reading from path " + path);
				} else {
					FileInputStream i = new FileInputStream(path);
					InputStreamReader r = new InputStreamReader(i, inputCharset);
					reader = new BufferedReader(r);
					
					if (logger.isTraceEnabled())
						logger.trace("Faucet has created reader " + reader +
								" wrapping reader " +r + " wrapping input stream " + i +
								" reading from path " + path);
				}
			} catch (Exception e) {
				throw new XformationException("Unable to create input stream", e);
			}
			break;
		case Reader:
			break;
		}
		
		return reader;
	}
	
	@Override
	public void dispose() {
		switch (type) {
		case File:
			try {
				reader.close();
			} catch (IOException e) {
				throw new XformationException("Unable to close input stream", e);
			}
			break;
		case Reader:
			// Caller owns the reader. It is their responsibility to close it.
			break;
		}
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		String sinkClassName = in.readUTF();
		Class<?> sinkClass;
		try {
			sinkClass = Class.forName(sinkClassName);
		} catch (ClassNotFoundException e) {
			throw new InstantiationException("Unable to load class: " + sinkClassName);
		}
		this.sink = (ISink) sinkClass.newInstance();
		this.sink.readFrom(in);
		this.hookedUp = true;
		this.sink.hookupFaucet(this);
		this.type = Type.File;
		this.reader = null;
		this.path = in.readUTF();
		this.inputCharset = in.readUTF();
	}

	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (type != Type.File)
			throw new XformationException(
					"Only ReaderFaucets of type Type.File can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(sink.getClass().getCanonicalName());
		sink.writeTo(out);
		out.writeUTF(path);
		out.writeUTF(inputCharset);
	}

	@Override
	public List<ISink> getTerminalSinks() {
		if (sink instanceof IFaucet)
			return ((IFaucet) sink).getTerminalSinks();
		else if (sink instanceof ISplitPipelineItem)
			return ((ISplitPipelineItem) sink).getTerminalSinks();
		else
			return Collections.singletonList(sink);
	}

}
