/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.pipeline;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.PipedReader;
import java.io.PipedWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Future;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.FaucetAdapter;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItem;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ExecutorsHelper;

public class SplitReaderPipelineItem implements ISplitPipelineItem {
	private static final Logger logger = Logger.getLogger(SplitReaderPipelineItem.class);
	
	protected String id;
	protected boolean hookedUpFaucet;
	protected List<Boolean> hookedUpSinkFlags;
	protected List<ISink> sinks;
	protected List<IFaucet> faucetAdapters;
	protected IFaucet faucet;
	protected Reader reader;
	protected List<PipedReader> readerForFaucets;
	protected List<PipedWriter> writers;
	protected Semaphore consumptionSemaphore;
	protected AtomicBoolean consumptionGate;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public SplitReaderPipelineItem() { }
	
	public SplitReaderPipelineItem(String id, int numberOfSinks) {
		this.id = id;
		this.hookedUpFaucet = false;
		this.hookedUpSinkFlags = new ArrayList<Boolean>(numberOfSinks);
		this.sinks = new ArrayList<ISink>(numberOfSinks);
		this.faucetAdapters = new ArrayList<IFaucet>(numberOfSinks);
		this.faucet = null;
		this.reader = null;
		this.readerForFaucets = new ArrayList<PipedReader>(numberOfSinks);
		this.writers = new ArrayList<PipedWriter>(numberOfSinks);
		this.consumptionSemaphore = new Semaphore(0);
		this.consumptionGate = new AtomicBoolean(true);
		
		for (int i = 0; i < numberOfSinks; i++) {
			this.hookedUpSinkFlags.add(false);
			this.sinks.add(null);
			this.faucetAdapters.add(null);
			this.readerForFaucets.add(null);
			this.writers.add(null);
		}
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public ISink getSink(int index) {
		return sinks.get(index);
	}

	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case CharStream:
			return true;
		case ByteStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case CharStream:
			return true;
		case ByteStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	protected void checkFaucetValidity(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.CharStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the CharStream content type.");
		if (faucet instanceof IDemuxFaucet)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IDemuxFaucet instance. This is not allowed.");
		if (faucet instanceof IFaucetTemplate)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IFaucetTemplate instance. This is not allowed.");
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		checkFaucetValidity(faucet);
		if (hookedUpFaucet && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpFaucet)
			throw new XformationException("Trying to hook up an already hooked up faucet.");
		
		this.faucet = faucet;
		this.hookedUpFaucet = true;
		
		this.faucet.hookupSink(this);
	}

	protected void checkSinkValidity(ISink sink) {
		if (!sink.supportsInput(ContentType.CharStream))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the CharStream content type.");
		if (sink instanceof IMuxSink)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an IMuxSink instance. This is not allowed.");
		if (sink instanceof ISinkTemplate)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an ISinkTemplate instance. This is not allowed.");
	}

	@Override
	public void hookupSink(int index, ISink sink) {
		checkSinkValidity(sink);
		if (hookedUpSinkFlags.get(index) && this.sinks.get(index) == sink) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpSinkFlags.get(index))
			throw new XformationException("Trying to hook up an already hooked up " +
					"sink at index " + index + ".");
		
		this.sinks.set(index, sink);
		this.hookedUpSinkFlags.set(index, true);
		
		this.faucetAdapters.set(index, new FaucetAdapter(this, index));
		this.sinks.get(index).hookupFaucet(this.faucetAdapters.get(index));
		
		this.consumptionSemaphore.release();
	}

	@Override
	public Object getSource(int index, ContentType contentType) {
		switch (contentType) {
		case ByteStream:
		case XML:
		case ResultSet:
			throw new XformationException("Content type: " + contentType +
			" is not supported.");
		}
		
		if (readerForFaucets.get(index) != null)
			return readerForFaucets.get(index);
		
		this.writers.set(index, new PipedWriter());
		try {
			this.readerForFaucets.set(
					index, new PipedReader(this.writers.get(index)));
		} catch (IOException e) {
			throw new XformationException("Unable to create hooked piped reader/writer", e);
		}
		
		if (logger.isTraceEnabled())
			logger.trace("Created for index " + index +
					", reader " + readerForFaucets.get(index) +
					" linked with writer " + writers.get(index));

		return readerForFaucets.get(index);
	}
	
	@Override
	public void registerSource(Object source) {
		// only used in mux/demux cases.
	}
	
	@Override
	public void handleMuxSwitch() {
		// only used in mux/demux cases.
	}
	
	@Override
	public void endMux() {
		// only used in mux/demux cases.
		
	}
	
	@Override
	public String dumpPipeline() {
		StringBuilder sb = new StringBuilder();
		int index = 0;
		for (ISink sink : sinks) {
			sb.append(sink.dumpPipeline() + "\n" + getClass().getName() + ": " + reader +
					"->" + writers.get(index) + "->" + readerForFaucets.get(index));
			index++;
		}
		return sb.toString();
	}
	
	@Override
	public void dispose(int index) {
		try {
			if (readerForFaucets.get(index) != null)
				readerForFaucets.get(index).close();
			readerForFaucets.set(index, null);
		} catch (IOException e) {
			throw new XformationException("Unable to close input stream", e);
		}
		
		boolean isLastDisposal = true;
		for (Reader reader : readerForFaucets) {
			if (reader != null) {
				isLastDisposal = false;
				break;
			}
		}
		
		if (isLastDisposal)
			faucet.dispose();
	}
	
	@Override
	public void dispose() {
		faucet.dispose();
	}
	
	@Override
	public void consume() {
		throw new XformationException("Calling ISink.consume() on a ISplitPipelineItem " +
				"implementation. Call ISplitPipelineItem.consume(int, ISink) instead.");
	}

	@Override
	public Future<?> consumeAsynchronously() {
		throw new XformationException("Calling ISink.consumeAsynchronously() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}

	@Override
	public Object consume(ISink previousSink) {
		if (!hookedUpFaucet)
			throw new XformationException("Pipeline item has not been set up " +
					"correctly: faucet has not been set");
		
		if (!consumptionGate.getAndSet(false))
			return null;
		
		consumptionSemaphore.acquireUninterruptibly(sinks.size());
		
	    if (faucet instanceof IPipelineItem)
	    	((IPipelineItem) faucet).consume(this);
	    
	    if (this instanceof ISinkTemplate) {
	    	if (!((ISinkTemplate) this).isClone()) {
	    		return null;
	    	}
	    }

	    boolean hookedUpSinks = true;
	    for (Boolean hookedUpSink : hookedUpSinkFlags) {
	    	if (!hookedUpSink) {
	    		hookedUpSinks = false;
	    		break;
	    	}
	    }
		if (!hookedUpSinks)
			throw new XformationException("Pipeline item has not been set up " +
					"correctly: sinks has not been all set");

	    reader = (Reader) faucet.getSource(ContentType.CharStream);
	    
	    for (int i = 0; i < sinks.size(); i++)
	    	getSource(i, ContentType.CharStream);

		ExecutorsHelper.getInstance().executeInProc(new Runnable() {
			@Override
			public void run() {
				try {
					int count;
					char[] buffer = new char[8 * 1024];
					if (logger.isTraceEnabled())
						logger.trace("Using reader " + reader + " for the stream transformation.");
					while ((count = reader.read(buffer)) != -1) {
						for (Writer writer : writers) {
							if (logger.isTraceEnabled()) {
								logger.trace("Read " + new String(buffer, 0, count) +
										" from reader " + reader +
										", writing to writer " + writer);
							}
							writer.write(buffer, 0, count);
						}
					}
					for (Writer writer : writers)
						writer.close();
					
					logger.debug("Executable for the stream pipeline will now exit.");
				} catch (Throwable t) {
					logger.error("Error during pipeline thread execution.", t);
				}
			}
		});
		
		return null;
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		int numberOfSinks = in.readInt();
		this.hookedUpFaucet = false;
		this.hookedUpSinkFlags = new ArrayList<Boolean>(numberOfSinks);
		this.sinks = new ArrayList<ISink>(numberOfSinks);
		this.faucetAdapters = new ArrayList<IFaucet>(numberOfSinks);
		this.faucet = null;
		this.reader = null;
		this.readerForFaucets = new ArrayList<PipedReader>(numberOfSinks);
		this.writers = new ArrayList<PipedWriter>(numberOfSinks);
		this.consumptionSemaphore = new Semaphore(0);
		this.consumptionGate = new AtomicBoolean(true);
		
		for (int i = 0; i < numberOfSinks; i++) {
			this.hookedUpSinkFlags.add(false);
			this.sinks.add(null);
			this.faucetAdapters.add(null);
			this.readerForFaucets.add(null);
			this.writers.add(null);
		}
		
		for (int i = 0; i < numberOfSinks; i++) {
			String sinkClassName = in.readUTF();
			Class<?> sinkClass;
			try {
				sinkClass = Class.forName(sinkClassName);
			} catch (ClassNotFoundException e) {
				throw new InstantiationException("Unable to load class: " + sinkClassName);
			}
			
			ISink sink = (ISink) sinkClass.newInstance();
			sink.readFrom(in);

			this.sinks.set(i, sink);
			this.hookedUpSinkFlags.set(i, true);
			
			this.faucetAdapters.set(i, new FaucetAdapter(this, i));
			this.sinks.get(i).hookupFaucet(this.faucetAdapters.get(i));
			
			this.consumptionSemaphore.release();

		}
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		out.writeUTF(id);
		out.writeInt(sinks.size());
		for (ISink s : sinks) {
			out.writeUTF(s.getClass().getCanonicalName());
			s.writeTo(out);
		}
	}

	@Override
	public List<ISink> getTerminalSinks() {
		List<ISink> retVal = new ArrayList<ISink>();
		for (ISink sink : sinks) {
			if (sink instanceof IFaucet)
				retVal.addAll(((IFaucet) sink).getTerminalSinks());
			else if (sink instanceof ISplitPipelineItem)
				retVal.addAll(((ISplitPipelineItem) sink).getTerminalSinks());
			else
				retVal.add(sink);
		}
		return retVal;
	}
}
