/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.pipeline;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.PipedReader;
import java.io.PipedWriter;
import java.util.ArrayList;
import java.util.UUID;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicBoolean;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.FaucetAdapter;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.IPipelineItemTemplate;
import com.koutra.dist.proc.model.ISimplePipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItemTemplate;
import com.koutra.dist.proc.model.XformationException;

public class SplitReaderPipelineItemTemplate extends SplitReaderPipelineItem
		implements ISplitPipelineItemTemplate {

	protected boolean isClone;

	public SplitReaderPipelineItemTemplate(int numberOfSinks) {
		super(UUID.randomUUID().toString(), numberOfSinks);
		isClone = false;
	}
	
	/**
	 * @deprecated use the initializing constructor instead.
	 */
	public SplitReaderPipelineItemTemplate() {
		super(UUID.randomUUID().toString(), 0);
		isClone = false;
	}
	
	@Override
	public boolean isClone() {
		return isClone;
	}

	@Override
	public void createClone(IPipelineItem pipelineItem) {
		try {
			SplitReaderPipelineItemTemplate clone = getClass().newInstance();
			int numberOfSinks = sinks.size();
			clone.id = id;
			clone.isClone = true;
			clone.hookedUpFaucet = false;
			clone.hookedUpSinkFlags = new ArrayList<Boolean>(numberOfSinks);
			clone.sinks = new ArrayList<ISink>(numberOfSinks);
			clone.faucetAdapters = new ArrayList<IFaucet>(numberOfSinks);
			clone.faucet = null;
			clone.reader = null;
			clone.readerForFaucets = new ArrayList<PipedReader>(numberOfSinks);
			clone.writers = new ArrayList<PipedWriter>(numberOfSinks);
			clone.consumptionSemaphore = new Semaphore(0);
			clone.consumptionGate = new AtomicBoolean(true);

			for (int i = 0; i < numberOfSinks; i++) {
				clone.hookedUpSinkFlags.add(false);
				clone.sinks.add(null);
				clone.faucetAdapters.add(null);
				clone.readerForFaucets.add(null);
				clone.writers.add(null);
			}
			
			if (pipelineItem instanceof ISimplePipelineItem)
				((ISimplePipelineItem) pipelineItem).hookupSink(clone);
			else if (pipelineItem instanceof FaucetAdapter)
				((FaucetAdapter) pipelineItem).hookupSink(clone);
			
			int index = 0;
			for (ISink sink : sinks) {
				if (sink instanceof ISinkTemplate) {
					FaucetAdapter adapter = new FaucetAdapter(clone, index++);
					((ISinkTemplate) sink).createClone(adapter);
					clone.sinks.set(index - 1, adapter.getSink());
				} else {
					sink.hookupFaucet(new FaucetAdapter(clone, index++));
				}
			}
		} catch (Exception e) {
			throw new XformationException("Unable to create clone", e);
		}
	}

	@Override
	protected void checkFaucetValidity(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.CharStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the CharStream content type.");
		if (!(faucet instanceof IDemuxFaucet) &&
				!(faucet instanceof IFaucetTemplate) &&
				!(faucet instanceof FaucetAdapter))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is neither an IDemuxFaucet nor an IFaucetTemplate instance, " +
					"nor a FaucetAdapter. " +
					"This is not allowed.");
	}

	@Override
	protected void checkSinkValidity(ISink sink) {
		if (!sink.supportsInput(ContentType.CharStream))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the CharStream content type.");
		if (!(sink instanceof IMuxSink) && !(sink instanceof ISinkTemplate))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is neither an IMuxSink or an ISinkTemplate instance. " +
					"This is not allowed.");
	}

	@Override
	public void handleMuxSwitch() {
		int index = 0;
		for (ISink sink : sinks) {
			if (sink instanceof IPipelineItem)
				((IPipelineItem) sink).handleMuxSwitch();
			else
				sink.registerSource(getSource(index++, ContentType.ByteStream));
		}
	}

	@Override
	public void endMux() {
		for (ISink sink : sinks) {
			if (sink instanceof IPipelineItemTemplate) {
				((IPipelineItemTemplate) sink).endMux();
			} else {
				sink.registerSource(null);
			}
		}
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		super.writeTo(out);
		out.writeBoolean(isClone);
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		super.readFrom(in);
		this.isClone = in.readBoolean();
	}
	
}
