/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.pipeline;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISimplePipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItem;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ByteArrayUtil;
import com.koutra.dist.proc.util.ExecutorsHelper;

/**
 * A pipeline item that reads information from a byte stream and writes to a byte
 * stream. Implementors need only give the implementation of the transformBuffer()
 * method.
 * 
 * @author Pafsanias Ftakas
 */
public abstract class StreamToStreamPipelineItem implements ISimplePipelineItem {
	private static final Logger logger = Logger.getLogger(StreamToStreamPipelineItem.class);
	
	protected String id;
	protected boolean hookedUpFaucet;
	protected boolean hookedUpSink;
	protected ISink sink;
	protected IFaucet faucet;
	protected InputStream inputStream;
	protected PipedInputStream inputStreamForFaucet;
	protected PipedOutputStream outputStream;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public StreamToStreamPipelineItem() { }
	
	public StreamToStreamPipelineItem(String id) {
		this.id = id;
		this.hookedUpFaucet = false;
		this.hookedUpSink = false;
		this.sink = null;
		this.faucet = null;
		this.inputStream = null;
		this.inputStreamForFaucet = null;
		this.outputStream = null;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public ISink getSink() {
		return sink;
	}

	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	protected void checkFaucetValidity(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.ByteStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the ByteStream content type.");
		if (faucet instanceof IDemuxFaucet)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IDemuxFaucet instance. This is not allowed.");
		if (faucet instanceof IFaucetTemplate)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IFaucetTemplate instance. This is not allowed.");
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		checkFaucetValidity(faucet);
		if (hookedUpFaucet && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpFaucet)
			throw new XformationException("Trying to hook up an already hooked up faucet.");
		
		this.faucet = faucet;
		this.hookedUpFaucet = true;
		
		this.faucet.hookupSink(this);
	}

	protected void checkSinkValidity(ISink sink) {
		if (!sink.supportsInput(ContentType.ByteStream))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the ByteStream content type.");
		if (sink instanceof IMuxSink)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an IMuxSink instance. This is not allowed.");
		if (sink instanceof ISinkTemplate)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an ISinkTemplate instance. This is not allowed.");
	}

	@Override
	public void hookupSink(ISink sink) {
		checkSinkValidity(sink);
		if (hookedUpSink && this.sink == sink) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpSink)
			throw new XformationException("Trying to hook up an already hooked up sink.");
		
		this.sink = sink;
		this.hookedUpSink = true;
		
		this.sink.hookupFaucet(this);
	}

	@Override
	public Object getSource(ContentType contentType) {
		switch (contentType) {
		case CharStream:
		case XML:
		case ResultSet:
			throw new XformationException("Content type: " + contentType +
			" is not supported.");
		}
		
		if (inputStreamForFaucet != null)
			return inputStreamForFaucet;
		
		this.outputStream = new PipedOutputStream();
		try {
			this.inputStreamForFaucet = new PipedInputStream(this.outputStream);
		} catch (IOException e) {
			throw new XformationException("Unable to create hooked piped streams", e);
		}
		
		if (logger.isTraceEnabled())
			logger.trace("Created input stream " + inputStreamForFaucet +
					" linked with output stream " + outputStream);

		return inputStreamForFaucet;
	}
	
	@Override
	public void registerSource(Object source) {
		// only used in mux/demux cases.
	}
	
	@Override
	public void handleMuxSwitch() {
		// only used in mux/demux cases.
	}
	
	@Override
	public void endMux() {
		// only used in mux/demux cases.
		
	}
	
	@Override
	public String dumpPipeline() {
		return sink.dumpPipeline() + "\n" + getClass().getName() + ": " + inputStream +
			"->" + outputStream + "->" + inputStreamForFaucet;
	}
	
	@Override
	public void dispose() {
		try {
			if (inputStreamForFaucet != null)
				inputStreamForFaucet.close();
		} catch (IOException e) {
			throw new XformationException("Unable to close input stream", e);
		}
		
		faucet.dispose();
	}
	
	/**
	 * This method is a callback that performs the transformation. The returned byte
	 * array will get written on the output stream that hooks this pipeline item to its
	 * sink down the pipeline chain.
	 * 
	 * @param buffer The buffer to read bytes from. Only the bytes in the specified
	 * range should be used.
	 * @param off The offset of the first byte in the buffer that is part of the range
	 * to use.
	 * @param len The number of bytes of the valid range, or null if the end of the
	 * input has been reached.
	 * @return The byte array to pass down the pipeline chain.
	 */
	public abstract byte[] transformBuffer(byte[] buffer, int off, int len);

	@Override
	public void consume() {
		throw new XformationException("Calling ISink.consume() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}

	@Override
	public Future<?> consumeAsynchronously() {
		throw new XformationException("Calling ISink.consumeAsynchronously() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}

	@Override
	public Object consume(ISink previousSink) {
		if (!hookedUpFaucet)
			throw new XformationException("Pipeline item has not been set up " +
					"correctly: faucet has not been set");
		
	    if (faucet instanceof IPipelineItem)
	    	((IPipelineItem) faucet).consume(this);
	    
	    if (this instanceof ISinkTemplate) {
	    	if (!((ISinkTemplate) this).isClone()) {
	    		return null;
	    	}
	    }

		if (!hookedUpSink)
			throw new XformationException("Pipeline item has not been set up " +
					"correctly: sink has not been set");

		inputStream = (InputStream) faucet.getSource(ContentType.ByteStream);
	    
	    getSource(ContentType.ByteStream);

		ExecutorsHelper.getInstance().executeInProc(new Runnable() {
			@Override
			public void run() {
				try {
					int count;
					byte[] buffer = new byte[8 * 1024];
					if (logger.isTraceEnabled())
						logger.trace("Using input stream " + inputStream + " for the stream transformation.");
					while ((count = inputStream.read(buffer)) != -1) {
						byte[] transformation = transformBuffer(buffer, 0, count);
						if (logger.isTraceEnabled()) {
							String readInput = ByteArrayUtil.getHexDump(buffer, 0, count);
							String transformedInput = ByteArrayUtil.getHexDump(transformation, 0, transformation.length);
							logger.trace("Read " + readInput + " from input stream " +
									inputStream + ", writing " + transformedInput +
									" to output stream " + outputStream);
						}
						outputStream.write(transformation, 0, transformation.length);
					}
					byte[] transformation = transformBuffer(buffer, 0, count);
					if (logger.isTraceEnabled()) {
						String transformedInput = ByteArrayUtil.getHexDump(transformation, 0, transformation.length);
						logger.trace("Writting final " + transformedInput +
								" to output stream " + outputStream);
					}
					outputStream.write(transformation, 0, transformation.length);
					outputStream.close();
					
					logger.debug("Executable for the stream pipeline will now exit.");
				} catch (Throwable t) {
					logger.error("Error during pipeline thread execution.", t);
				}
			}
		});
		
		return null;
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		String sinkClassName = in.readUTF();
		Class<?> sinkClass;
		try {
			sinkClass = Class.forName(sinkClassName);
		} catch (ClassNotFoundException e) {
			throw new InstantiationException("Unable to load class: " + sinkClassName);
		}
		this.sink = (ISink) sinkClass.newInstance();
		this.sink.readFrom(in);
		this.hookedUpSink = true;
		this.sink.hookupFaucet(this);
		this.faucet = null;
		this.hookedUpFaucet = false;
		this.inputStream = null;
		this.inputStreamForFaucet = null;
		this.outputStream = null;
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		out.writeUTF(id);
		out.writeUTF(sink.getClass().getCanonicalName());
		sink.writeTo(out);
	}

	@Override
	public List<ISink> getTerminalSinks() {
		if (sink instanceof IFaucet)
			return ((IFaucet) sink).getTerminalSinks();
		else if (sink instanceof ISplitPipelineItem)
			return ((ISplitPipelineItem) sink).getTerminalSinks();
		else
			return Collections.singletonList(sink);
	}

}
